package java.util.concurrent.atomic;

import java.util.function.IntBinaryOperator;
import java.util.function.IntUnaryOperator;
import sun.misc.Unsafe;
import java.io.Serializable;

public class AtomicInteger extends Number implements Serializable
{
    private static final long serialVersionUID = 6214790243416807050L;
    private static final Unsafe unsafe;
    private static final long valueOffset;
    private volatile int value;
    
    public AtomicInteger(int value) {
        this.value = value;
    }
    
    public AtomicInteger() {
    }
    
    public final int get() {
        return this.value;
    }
    
    public final void set(int value) {
        this.value = value;
    }
    
    public final void lazySet(int x) {
        AtomicInteger.unsafe.putOrderedInt(this, AtomicInteger.valueOffset, x);
    }
    
    public final int getAndSet(int newValue) {
        return AtomicInteger.unsafe.getAndSetInt(this, AtomicInteger.valueOffset, newValue);
    }
    
    public final boolean compareAndSet(int expected, int x) {
        return AtomicInteger.unsafe.compareAndSwapInt(this, AtomicInteger.valueOffset, expected, x);
    }
    
    public final boolean weakCompareAndSet(int expected, int x) {
        return AtomicInteger.unsafe.compareAndSwapInt(this, AtomicInteger.valueOffset, expected, x);
    }
    
    public final int getAndIncrement() {
        return AtomicInteger.unsafe.getAndAddInt(this, AtomicInteger.valueOffset, 1);
    }
    
    public final int getAndDecrement() {
        return AtomicInteger.unsafe.getAndAddInt(this, AtomicInteger.valueOffset, -1);
    }
    
    public final int getAndAdd(int delta) {
        return AtomicInteger.unsafe.getAndAddInt(this, AtomicInteger.valueOffset, delta);
    }
    
    public final int incrementAndGet() {
        return AtomicInteger.unsafe.getAndAddInt(this, AtomicInteger.valueOffset, 1) + 1;
    }
    
    public final int decrementAndGet() {
        return AtomicInteger.unsafe.getAndAddInt(this, AtomicInteger.valueOffset, -1) - 1;
    }
    
    public final int addAndGet(int delta) {
        return AtomicInteger.unsafe.getAndAddInt(this, AtomicInteger.valueOffset, delta) + delta;
    }
    
    public final int getAndUpdate(IntUnaryOperator intUnaryOperator) {
        int value;
        do {
            value = this.get();
        } while (!this.compareAndSet(value, intUnaryOperator.applyAsInt(value)));
        return value;
    }
    
    public final int updateAndGet(IntUnaryOperator intUnaryOperator) {
        int value;
        int applyAsInt;
        do {
            value = this.get();
            applyAsInt = intUnaryOperator.applyAsInt(value);
        } while (!this.compareAndSet(value, applyAsInt));
        return applyAsInt;
    }
    
    public final int getAndAccumulate(int n, IntBinaryOperator intBinaryOperator) {
        int value;
        do {
            value = this.get();
        } while (!this.compareAndSet(value, intBinaryOperator.applyAsInt(value, n)));
        return value;
    }
    
    public final int accumulateAndGet(int n, IntBinaryOperator intBinaryOperator) {
        int value;
        int applyAsInt;
        do {
            value = this.get();
            applyAsInt = intBinaryOperator.applyAsInt(value, n);
        } while (!this.compareAndSet(value, applyAsInt));
        return applyAsInt;
    }
    
    @Override
    public String toString() {
        return Integer.toString(this.get());
    }
    
    @Override
    public int intValue() {
        return this.get();
    }
    
    @Override
    public long longValue() {
        return (long)this.get();
    }
    
    @Override
    public float floatValue() {
        return (float)this.get();
    }
    
    @Override
    public double doubleValue() {
        return (double)this.get();
    }
    
    static {
        unsafe = Unsafe.getUnsafe();
        try {
            valueOffset = AtomicInteger.unsafe.objectFieldOffset(AtomicInteger.class.getDeclaredField("value"));
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
    }
}
